from ..compilation_error import CompilationError
from .kwargs_parser import parse_kwargs, apply_aliases

ARGS = ("find", "group", "enemy", "ore", "outX", "outY", "found", "building")
ARG_ALIASES = {
    "type": "find",
    "isEnemy": "enemy",
    "oreType": "ore",
    "resultX": "outX",
    "resultY": "outY",
    "resultFound": "found",
    "resultIsFound": "found",
    "resultBuilding": "building",
}

def unit_locate(src_line: str) -> list:
    """Enhanced unit locate instruction, with Python-style kwargs.
    Example:
    unit-locate type=ore oreType=@coal resultX=x resultY=y resultIsFound=found
    unit-locate type=building group=core isEnemy=false outX=x outY=y found=found building=core
    # Aliases
    unit-locate find=building group=core enemy=false outX=x outY=y found=found building=core
    unit-locate type=spawn resultX=x resultY=y resultIsFound=found building=building
    unit-locate type=damaged outX=x outY=y resultIsFound=found resultBuilding=building
    """
    verdicts = src_line.split()
    kwargs = parse_kwargs(verdicts[1:])
    kwargs = apply_aliases(kwargs, ARG_ALIASES)

    if kwargs["find"] == "ore":
        kwargs["group"] = "core"
        kwargs["enemy"] = "false"
        kwargs["building"] = "0"
    elif kwargs["find"] == "building":
        kwargs["ore"] = "0"
    elif kwargs["find"] == "spawn" or kwargs["find"] == "damaged":
        kwargs["group"] = "core"
        kwargs["enemy"] = "false"
        kwargs["ore"] = "0"
    else:
        message = F"error: type '{kwargs['find']}' is not supported"
        raise CompilationError(message)

    output_verdicts = ["ulocate", ]
    try:
        for arg in ARGS:
            output_verdicts.append(kwargs[arg])
    except KeyError as exception:
        name = exception.args[0]
        message = F"error: argument '{name}' is required, but not found"
        raise CompilationError(message) from exception
    return [" ".join(output_verdicts), ]



from .compilation_error import CompilationError

class BasicCompiler:
    """Basic compiler, supports xjump instruction."""
    def __init__(self):
        self.external_instructions = {}

    def add_instruction(self, instruction: str, handler):
        if instruction.startswith(":"):
            raise ValueError("Instructions that start with ':' is disallowed.")
        self.external_instructions[instruction] = handler

    def compile(self, src_text: str, sep='\n') -> str:
        code_lines = src_text.splitlines()
        code_lines = self.convert_externals(code_lines)

        code_lines, tags = parse_tags(code_lines)
        code_lines = convert_xjumps(code_lines, tags)
        code_lines = replace_tag_counter_macros(code_lines, tags)
        return sep.join(code_lines) + sep

    def convert_externals(self, src_lines: list) -> list:
        """Convert added instructions, remove empty lines, etc."""
        dst_lines = []
        for (src_cursor, src_line) in enumerate(src_lines):
            src_line = src_line.strip().rstrip()
            try:
                verdicts = src_line.split()
                if verdicts[0] in self.external_instructions.keys():
                    handler = self.external_instructions[verdicts[0]]
                    dst_lines.extend(handler(src_line))
                else:
                    # Pass them as-is
                    # They can be vanilla-mlog instructions ,xjump or tag
                    dst_lines.append(src_line)
            except IndexError:
                # verdicts[0]
                pass
            except CompilationError as exception:
                ext_info = ""
                if len(exception.args) > 0:
                    ext_info = " ".join(exception.args)

                message = F"line {src_cursor+1}: {ext_info}"
                raise CompilationError(message) from exception
        return dst_lines

def convert_xjumps(src_lines: list, dst_tagged: dict) -> list:
    dst_lines = []

    for (src_cursor, src_line) in enumerate(src_lines):
        try:
            verdicts = src_line.split()
            if verdicts[0] == "xjump":
                tag_name = verdicts[1]
                real_destination = dst_tagged[tag_name]

                # Mindustry logic instruction
                verdicts[0] = "jump"
                verdicts[1] = str(real_destination)

                dst_lines.append(" ".join(verdicts))
            else:
                dst_lines.append(src_line)
        except KeyError as exception:
            if len(exception.args) >= 1:
                message = F"line {src_cursor+1}: error: No such tag '{exception.args[0]}'"
                raise CompilationError(message) from exception
            raise exception

    return dst_lines

def parse_tags(src_lines: list) -> tuple:
    """Parse tags.
    If there are any tags at the end, an no-op instruction will be added.
    """
    dst_tagged = {}
    dst_lines = []

    dst_cursor = 0
    last_tagged_line = 0

    for src_line in src_lines:
        try:
            # One line, one tag
            # But you can apply multiple tags on one destination line
            verdicts = src_line.split()
            if verdicts[0].startswith(":"):
                tag_name = verdicts[0][1:]
                dst_tagged[tag_name] = dst_cursor
                last_tagged_line = dst_cursor
            else:
                dst_lines.append(src_line.lstrip().rstrip())
                dst_cursor += 1
        except IndexError:
            # Possibly empty lines
            pass

    if last_tagged_line == len(dst_lines):
        dst_lines.append("end")
    return (dst_lines, dst_tagged)

def replace_tag_counter_macros(src_lines: list, tags: dict) -> list:
    """Trivial __TAG_COUNTER macro processor.
Replace __TAG_COUNTER(tag_name) with corresponding destination code @counter.
"""
    result = []
    MACRO_HEADER = "__TAG_COUNTER("
    MACRO_HEADER_LEN = len(MACRO_HEADER)
    for (src_cursor, src_line) in enumerate(src_lines):
        verdicts = src_line.split()
        new_verdicts = []
        for arg in verdicts:
            if not arg.startswith(MACRO_HEADER) or not arg.endswith(")"):
                new_verdicts.append(arg)
                continue
            tag_name = arg[MACRO_HEADER_LEN:-1]
            try:
                tag_line = tags[tag_name]
                new_verdicts.append(str(tag_line))
            except KeyError as exception:
                if len(exception.args) >= 1:
                    message = F"line {src_cursor+1}: error: No such tag '{exception.args[0]}'"
                    raise CompilationError(message) from exception
                raise exception
        result.append(" ".join(new_verdicts))
    return result



from .extra_instructions.jump_if import COMPARATORS
from .extra_instructions.extended_let import BINARY_OPERATORS
from .extra_instructions.extended_let import BINARY_ASSIGNERS
from .extra_instructions.extended_let import UNARY_ASSIGNERS

DECOMPILER_TAG_PREFIX = "L"

class BasicDecompiler:
    def __init__(self):
        self.line_tags = {}
        self.jump_if_mapping = get_reversed_mapping(COMPARATORS)
        self.xlet_convertor = XletConvertor()

    def decompile(self, src_text:str, sep='\n') -> str:
        src_lines = src_text.splitlines()
        dst_lines = []
        self._parse_jumps(src_lines)
        for line_number, src_line in enumerate(src_lines):
            if line_number in self.line_tags.keys():
                tag_name = ":" + self.line_tags[line_number]
                dst_lines.append(tag_name)
            instruction = src_line.split(None, 1)[0]
            if instruction == "jump":
                dst_lines.append(self._convert_jump(src_line))
            elif instruction in self.xlet_convertor.handlers.keys():
                dst_lines.append(self.xlet_convertor.convert(src_line))
            else:
                dst_lines.append(src_line)
        return sep.join(dst_lines)

    def compile(self, src_text:str, sep='\n') -> str:
        return self.decompile(src_text, sep)

    def _parse_jumps(self, src_lines:list) -> None:
        for src_line in src_lines:
            verdicts = src_line.split()
            if len(verdicts) == 0:
                continue
            if verdicts[0] == "jump":
                tag_name = DECOMPILER_TAG_PREFIX + verdicts[1]
                line_number = int(verdicts[1])
                self.line_tags[line_number] = tag_name

    def _convert_jump(self, src_line:str) -> str:
        verdicts = src_line.split()
        dst_line_number = int(verdicts[1])
        dst_tag = self.line_tags[dst_line_number]
        comparator = self.jump_if_mapping[verdicts[2]]
        if comparator == "always":
            return F"jump-if {dst_tag} always"
        return F"jump-if {dst_tag} {verdicts[3]} {comparator} {verdicts[4]}"

class XletConvertor:
    """Convert op, set, sensor and getlink to xlet."""
    xlet = "xlet"
    def __init__(self):
        self.binary_assigners = get_reversed_mapping(BINARY_ASSIGNERS)
        self.unary_assigners = get_reversed_mapping(UNARY_ASSIGNERS)
        self.binary_operators = get_reversed_mapping(BINARY_OPERATORS)
        self.handlers = {
            "op": self.convert_op,
            "set": self.convert_set,
            "getlink": self.convert_getlink,
            "sensor": self.convert_sensor,
        }

    def convert(self, src_line: str) -> str:
        verdicts = src_line.split()
        result = ""
        try:
            handle = self.handlers[verdicts[0]]
            result = handle(verdicts)
        except IndexError:
            pass
        return result

    def convert_op(self, verdicts: list) -> str:
        action, destination = verdicts[1:3]
        value_a, value_b = verdicts[3:5]
        if action in self.binary_assigners.keys():
            xlet_action = self.binary_assigners[action]
            return F"xlet {destination} ={xlet_action} {value_a} {value_b}"
        if action in self.binary_operators.keys():
            xlet_action = self.binary_operators[action]
            return F"xlet {destination} = {value_a} {xlet_action} {value_b}"
        if action in self.unary_assigners.keys():
            xlet_action = self.unary_assigners[action]
            return F"xlet {destination} ={xlet_action} {value_a}"
        return " ".join(verdicts)

    def convert_set(self, verdicts: list) -> str:
        lvalue, rvalue = verdicts[1:3]
        return F"xlet {lvalue} = {rvalue}"

    def convert_getlink(self, verdicts: list) -> str:
        lvalue, link_number = verdicts[1:3]
        return F"xlet {lvalue} =getlink {link_number}"

    def convert_sensor(self, verdicts: list) -> str:
        lvalue, being_sensed, attribute = verdicts[1:4]
        return F"xlet {lvalue} =sensor {being_sensed} {attribute}"

def get_reversed_mapping(raw_mapping: dict) -> dict:
    """Get "reversed" mapping, work on dictionaries."""
    reversed_mapping = {}
    for key, value in raw_mapping.items():
        reversed_mapping[value] = key
    return reversed_mapping



import argparse
import sys
from . import BasicCompiler
from . import ExtendedCompiler
from . import ProceduralCompiler
from . import BasicDecompiler

parser = argparse.ArgumentParser(description="Extended Minductry Logic Compiler.",
        epilog="- stands for standard input/output")
parser.add_argument('input_file', type=str,
        help="Input file, like source code.")
parser.add_argument('output_file', type=str,
        help="Output file, usually vanilla mindustry logic code.")
parser.add_argument('--basic', '-b', action='store_true',
        help="Use basic compiler only (xjump instruction)")
parser.add_argument('--extended', '-e', action='store_true',
        help="Use extended compiler")
parser.add_argument('--procedural', '-p', action='store_true',
        help="Use procedural compiler (default)")
parser.add_argument('--dump', '-D', action='store_true',
        help="For procedural compiler, process if-else and while-wend only.")
parser.add_argument('--decompile', '-d', action='store_true',
        help="Decompile vanilla Mindustry logic to MlogExtended code")


def openfile(filename: str, mode: str, default=None):
    if filename == "-":
        if default is None:
            raise ValueError
        return default
    return open(filename, mode, encoding="utf-8")

if __name__ == '__main__':
    args = parser.parse_args()
    CompilerClass = ProceduralCompiler
    if args.basic:
        CompilerClass = BasicCompiler
    elif args.extended:
        CompilerClass = ExtendedCompiler
    elif args.decompile:
        CompilerClass = BasicDecompiler
    compiler = CompilerClass()

    with openfile(args.input_file, 'r', sys.stdin) as f_in:
        source_code = f_in.read()
        result = ""
        if args.dump:
            result = compiler.compile_to_backend(source_code)
        else:
            result = compiler.compile(source_code)
        with openfile(args.output_file, 'w', sys.stdout) as f_out:
            f_out.write(result)



"""Maintain jump tags for if-elif-else branches and while loops."""

class WhileLoopFlags:
    """Maintain while loop flags."""
    def __init__(self, identifier: str):
        self.tag_prefix = F"__mlogex_while_loop_{identifier}"
        self.while_looper = []

    def get_condition_tag(self):
        return F"{self.tag_prefix}_loop_cond"

    def get_loopbody_start_tag(self):
        return F"{self.tag_prefix}_loop_start"

    def get_loopbody_end_tag(self):
        return F"{self.tag_prefix}_loop_end"


class IfElseFlags:
    """Main if-elif-else branches and flags."""
    def __init__(self, identifier: str):
        self.tag_prefix = F"__mlogex_branches_{identifier}"
        self.branch_id = 0

    def get_current_branch_tag(self):
        return F"{self.tag_prefix}_{self.branch_id}"

    def get_next_branch_tag(self):
        return F"{self.tag_prefix}_{self.branch_id+1}"

    def get_endif_tag(self):
        return F"{self.tag_prefix}_endif"

    def increase_branch_id(self):
        self.branch_id += 1



from ..compilation_error import CompilationError

# From Mindustry 126.2
COMPARATORS = {
    "===": "strictEqual",
    "==": "equal",
    "!=": "notEqual",
    "<": "lessThan",
    "<=": "lessThanEq",
    ">": "greaterThan",
    ">=": "greaterThanEq",
    "always": "always"
}

def jump_if(src_line: str) -> list:
    """Usage:
    jump-if <Tag> always
    jump-if <Tag> x <= 42
    """
    result = []
    verdicts = src_line.split()
    # Supported by BasicCompiler
    verdicts[0] = "xjump"

    if len(verdicts) == 3:
        if verdicts[2] == "always":
            verdicts.extend(("0", "0"))
            line = " ".join(verdicts)
            result.append(line)
            return result
        raise CompilationError("error: too few arguments for jump-if")
    if len(verdicts) < 3:
        raise CompilationError("error: too few arguments for jump-if")
    if len(verdicts) == 4:
        raise CompilationError("error: invalid or incomplete arguments for jump-if")
    if len(verdicts) > 5:
        raise CompilationError("error: too many arguments for jump-if")

    (verdicts[2], verdicts[3]) = (verdicts[3], verdicts[2])
    try:
        verdicts[2] = COMPARATORS[verdicts[2]]
    except KeyError as exception:
        raise CompilationError(F"error: '{verdicts[2]}' is not supported") from  exception
    line = " ".join(verdicts)
    result.append(line)
    return result



from ..compilation_error import CompilationError
from .kwargs_parser import parse_kwargs, pad_args

TOTAL_VERDICTS = 7

COMMAND_ARGS = {
    "idle": (),
    "stop": (),
    "move": ("x", "y"),
    "approach": ("x", "y", "radius"),
    "boost": ("enable", ),
    "pathfind": (),
    "target": ("x", "y", "shoot"),
    "targetp": ("unit", "shoot"),
    "itemDrop": ("to", "amount"),
    "itemTake": ("from", "item", "amount"),
    "payDrop": (),
    "payTake": ("takeUnits", ),
    "mine": ("x", "y"),
    "flag": ("value", ),
    "build": ("x", "y", "block", "rotation", "config"),
    "getBlock": ("x", "y", "type", "building"),
    "within": ("x", "y", "radius", "result"),
}

COMMAND_ARG_ALIASES = {
    "boost": {
        "boost": "enable",
    },
    "targetp": {
        "target": "unit",
    },
    "flag": {
        "flag": "value",
    },
    "getBlock": {
        "resultType": "type",
        "resultBuilding": "building",
    },
}

def unit_control(src_line: str) -> list:
    """Enhanced unit control instruction, with Python-style kwargs.
    """
    verdicts = src_line.split()
    command = verdicts[1]
    kwargs = parse_kwargs(verdicts[2:])
    result = []

    try:
        if command not in COMMAND_ARGS.keys():
            message = F"error: unsupported unit command '{command}'"
            raise CompilationError(message)

        vanilla_kwargs = []
        if command in COMMAND_ARG_ALIASES.keys():
            aliases = COMMAND_ARG_ALIASES[command]
            for alias in aliases.keys():
                if alias in kwargs.keys():
                    vanilla_kwargs.append((aliases[alias], kwargs[alias]))
        for (k, v) in vanilla_kwargs:
            kwargs[k] = v

        output_verdicts = []
        output_verdicts.append("ucontrol")
        output_verdicts.append(command)
        for key in COMMAND_ARGS[command]:
            output_verdicts.append(kwargs[key])
        output_verdicts = pad_args(output_verdicts, TOTAL_VERDICTS)
        result.append(" ".join(output_verdicts))
    except KeyError as exception:
        name = ""
        if len(exception.args) >= 1:
            name = exception.args[0]
        message = F"error: argument '{name}' is required, but not found"
        raise CompilationError(message) from exception

    return result



"""Vanilla Mindustry logic support "function calling" by setting @counter.
However, vanilla Mindustry logic has neither stack nor namespace.
"Functions" may accidentally overwrite variables outside of them.
This module provides a thin wrapper for function calls ans returns.
"""
from ..compilation_error import CompilationError

def unsafe_call(src_line: str) -> list:
    """Provide __unsafe_call instruction. Will assign a variable to store return address.
    DOES NOT SUPPORT RECURSION!
    Usage:
    __unsafe_call SomeTag
    Note:
    The instruction above assigns an variable "SomeTag_return_address"
    """
    result = []
    verdicts = src_line.split()

    if len(verdicts) < 2:
        raise CompilationError("error: too few arguments for __unsafe_call")
    if len(verdicts) > 2:
        raise CompilationError("error: too many arguments for __unsafe_call")

    target_tag = verdicts[1]
    return_address_variable = target_tag + "_return_address"
    result.append(F"op add {return_address_variable} @counter 1")
    result.append(F"xjump {target_tag} always 0 0")
    return result

def unsafe_return(src_line: str) -> list:
    """Provide __unsafe_return instruction. Return to previously assigned return address.
    Usage:
    __unsafe_return SomeTag
    Note:
    The instruction above overwrites @counter with the value of SomeTag_return_address.
    That is, probably return to the callee.
    """
    result = []
    verdicts = src_line.split()

    if len(verdicts) < 2:
        raise CompilationError("error: too few arguments for __unsafe_return")
    if len(verdicts) > 2:
        raise CompilationError("error: too many arguments for __unsafe_return")

    target_tag = verdicts[1]
    return_address_variable = target_tag + "_return_address"
    result.append(F"set @counter {return_address_variable}")
    return result



"""An Extended basic compiler."""
from .basic_compiler import BasicCompiler
from .extra_instructions import INSTRUCTIONS

class ExtendedCompiler(BasicCompiler):
    """Similar to BasicCompiler, but support more instructions."""
    def __init__(self):
        super().__init__()
        for (instruction, handler) in INSTRUCTIONS.items():
            self.add_instruction(instruction, handler)



"""Procedural compiler module.
Provide ProceduralCompiler."""
from .compilation_error import CompilationError
from .extended_compiler import ExtendedCompiler
from .procedural_structures import WhileLoopFlags, IfElseFlags

INVERT_TABLE = {
    "!=": "==",
    "==": "!=",
    "<": ">=",
    "<=": ">",
    ">": "<=",
    ">=": "<",
}

class ProceduralCompiler:
    """Procedural compiler, provides procedural-programming-like experience.
if, elif and while can receive 2 arguments at most.
e.g.:
    if a && b
    elif b < 0
    elif a
    else
    while i > 0
Keywords:
    if: starts the first branch
    elif: involved if there is more than 1 branch.
    else: indicates the last branch
    endif: terminates an if-else branch, necessary.

    while: starts a while loop.
    endwhile: ends a while loop.
    wend: same as "endwhile"
    """

    if_keywords = ("if", "elif", "else", "endif")
    while_keywords = ("while", "wend", "endwhile", "break", "continue")

    def __init__(self):
        self.backend = ExtendedCompiler()
        self.loop_stack = []
        self.if_stack = []

    def compile_to_backend(self, src_text:str, sep='\n') -> str:
        dst_lines = []
        for (line_number, src_line) in enumerate(src_text.splitlines()):
            verdicts = src_line.split()
            if len(verdicts) == 0:
                continue
            if verdicts[0] in self.if_keywords:
                try:
                    dst_lines.extend(self.handle_if(verdicts, str(line_number)))
                except IndexError as exception:
                    message = F"line {line_number+1}: error: "
                    message += F"{verdicts[0]} without if."
                    raise CompilationError(message) from exception
            elif verdicts[0] in self.while_keywords:
                try:
                    dst_lines.extend(self.handle_while(verdicts, str(line_number)))
                except IndexError as exception:
                    message = F"line {line_number+1}: error: "
                    message += F"{verdicts[0]} outside of loop."
                    raise CompilationError(message) from exception
            else:
                dst_lines.append(" ".join(verdicts))
        return sep.join(dst_lines)

    def handle_if(self, verdicts:list, identifier: str) -> list:
        result = []
        condition = "__mlogex_ifelse_condition"
        if verdicts[0] == "if":
            self.if_stack.append(IfElseFlags(identifier))
            top_tagger = self.if_stack[-1]
            next_tag = top_tagger.get_next_branch_tag()
            top_tagger.increase_branch_id()
            jump_instructions = get_inverted_jump(condition, next_tag, verdicts)
            result.extend(jump_instructions)
            return result
        top_tagger = self.if_stack[-1]
        if verdicts[0] in ("elif", "else"):
            current_tag = top_tagger.get_current_branch_tag()
            next_tag = top_tagger.get_next_branch_tag()
            final_tag = top_tagger.get_endif_tag()
            top_tagger.increase_branch_id()
            result.append(F"jump-if {final_tag} always")
            result.append(F":{current_tag}")
            if verdicts[0] == "elif":
                jump_instructions = get_inverted_jump(condition, next_tag, verdicts)
                result.extend(jump_instructions)
            return result
        if verdicts[0] == "endif":
            current_tag = top_tagger.get_current_branch_tag()
            final_tag = top_tagger.get_endif_tag()
            result.append(F":{current_tag}")
            result.append(F":{final_tag}")
            self.if_stack.pop()
            return result
        return []

    def handle_while(self, verdicts:list, identifier: str) -> list:
        result = []
        condition = "__mlogex_while_condition"
        if verdicts[0] == "while":
            self.loop_stack.append(WhileLoopFlags(identifier))
            top_tagger = self.loop_stack[-1]
            start_tag = top_tagger.get_loopbody_start_tag()
            cond_tag = top_tagger.get_condition_tag()

            top_tagger.while_looper.append(F":{cond_tag}")
            jumps = get_fastest_while_jump(condition, start_tag, verdicts)
            top_tagger.while_looper.extend(jumps)
            result.append(F"jump-if {cond_tag} always")
            result.append(F":{start_tag}")
            return result
        top_tagger = self.loop_stack[-1]
        cond_tag = top_tagger.get_condition_tag()
        end_tag = top_tagger.get_loopbody_end_tag()
        if verdicts[0] in ("wend", "endwhile"):
            result.extend(top_tagger.while_looper)
            # result.append(F"# {repr(top_tagger.while_looper)}")
            # result.append(F"# {repr(verdicts)}")
            result.append(F":{end_tag}")
            self.loop_stack.pop()
        elif verdicts[0] == "break":
            result.append(F"jump-if {end_tag} always")
        elif verdicts[0] == "continue":
            result.append(F"jump-if {cond_tag} always")
        return result

    def compile(self, src_text:str, sep='\n') -> str:
        extended_source_code = self.compile_to_backend(src_text, sep)
        return self.backend.compile(extended_source_code, sep)

def create_temporary_xlet(variable: str, arguments: list) -> str:
    return F"xlet {variable} = " + (" ".join(arguments))

def try_invert_xlet(xlet_instruction: str, target_tag: str) -> str:
    """Optimize performance for if-elif and while-wend control flows.
    If operator is invertable (e.g.: <, >, <=, >=, ==, !=), return inverted expression.
    If it is not invertable, return empty string.
    Example:
    "xlet condition = a < b", "tag" -> "jump-if tag a >= b"
    "xlet condition = a && b", "tag" -> ""
    """
    verdicts = xlet_instruction.split()
    if len(verdicts) == 4:
        real_condition = verdicts[3]
        return F"jump-if {target_tag} {real_condition} == false"
    operator = verdicts[4]
    if operator in INVERT_TABLE.keys():
        inverted = INVERT_TABLE[operator]
        return F"jump-if {target_tag} {verdicts[3]} {inverted} {verdicts[5]}"
    return ""

def get_inverted_jump(condition:str, jump_tag:str, verdicts: list) -> list:
    result = []
    temp_xlet = create_temporary_xlet(condition, verdicts[1:])
    inverted = try_invert_xlet(temp_xlet, jump_tag)
    if inverted:
        result.append(inverted)
    else:
        result.append(temp_xlet)
        result.append(F"jump-if {jump_tag} {condition} == false")
    return result

def get_fastest_while_jump(condition:str, jump_tag:str, verdicts: list) -> list:
    """Verdicts like ["while", "a", "<", "10"] """
    result = []
    jumpables = ("===", ) + tuple(INVERT_TABLE.keys())
    if len(verdicts) == 2:
        result.append(F"jump-if {jump_tag} {verdicts[1]} != false")
    elif verdicts[2] in jumpables:
        result.append(F"jump-if {jump_tag} " + (" ".join(verdicts[1:]) ) )
    else:
        result.append(create_temporary_xlet(condition, verdicts[1:]))
        result.append(F"jump-if {jump_tag} {condition} != false")
    return result



from ..compilation_error import CompilationError
from .kwargs_parser import parse_kwargs, apply_aliases, pad_args

TOTAL_VERDICTS = 7

ACTION_ALIASES = {
    "toggle": "enabled",
    "config": "configure",
}

ACTION_ARGS = {
    "enabled": ("status", ),
    "shoot": ("x", "y", "shoot"),
    "shootp": ("unit", "shoot"),
    "configure": ("config", ),
    "color": ("r", "g", "b"),
}

ACTION_ARG_ALIASES = {
    "shootp": {
        "target": "unit",
    },
}

def control(src_line: str) -> list:
    """Enhanced control instruction, with Python-style kwargs.
    Examples:
    # Disable a generator
    xcontrol generator1 action=toggle status=0
    xcontrol generator1 action=enabled status=0
    # Control a cyclone (turret), using argument aliases(unit vs target)
    xcontrol cyclone1 action=shoot x=enemyX y=enemyY shoot=1
    xcontrol cyclone1 action=shoot x=enemyX y=enemyY shoot=0
    xcontrol cyclone1 action=shootp unit=enemy shoot=1
    xcontrol cyclone1 action=shootp target=enemy shoot=1
    # Config a sorter to sort different items
    xcontrol sorter1 action=configure config=@copper
    xcontrol sorter1 action=config config=@lead
    # Set illuminator's color
    xcontrol illuminator1 action=color r=255 g=153 b=0
    """
    verdicts = src_line.split()
    target_building = verdicts[1]
    kwargs = parse_kwargs(verdicts[2:])

    result = []

    try:
        action = kwargs["action"]
        if action in ACTION_ALIASES.keys():
            action = ACTION_ALIASES[action]
        if action not in ACTION_ARGS.keys():
            message = F"error: unsupported control action '{action}'"
            raise CompilationError(message)
        if action in ACTION_ARG_ALIASES.keys():
            kwargs = apply_aliases(kwargs, ACTION_ARG_ALIASES[action])

        output_verdicts = []
        output_verdicts.append("control")
        output_verdicts.append(action)
        output_verdicts.append(target_building)
        for key in ACTION_ARGS[action]:
            output_verdicts.append(kwargs[key])
        output_verdicts = pad_args(output_verdicts, TOTAL_VERDICTS)
        result.append(" ".join(output_verdicts))
    except KeyError as exception:
        name = ""
        if len(exception.args) >= 1:
            name = exception.args[0]
        message = F"error: argument '{name}' is required, but not found"
        raise CompilationError(message) from exception

    return result



"""Mindustry Logic Extended: Provide extra instructions for mindustry logic.
"""
from .compilation_error import CompilationError
from .basic_compiler import BasicCompiler
from .extended_compiler import ExtendedCompiler
from .procedural_compiler import ProceduralCompiler
from .basic_decompiler import BasicDecompiler



from ..compilation_error import CompilationError
from .kwargs_parser import parse_kwargs, apply_aliases, pad_args

TOTAL_VERDICTS = 8

ACTION_ALIASES = {
    "rectangle": "rect",
    "lineRectangle": "lineRect",
    "polygon": "poly",
    "linePolygon": "linePoly",
}

ACTION_ARGS = {
    "clear": ("r", "g", "b"),
    "stroke": ("width", ),
    "color": ("r", "g", "b"),
    "line": ("x", "y", "x2", "y2"),
    "rect": ("x", "y", "width", "height"),
    "lineRect": ("x", "y", "width", "height"),
    "poly": ("x", "y", "sides", "radius", "rotation"),
    "linePoly": ("x", "y", "sides", "radius", "rotation"),
    "image": ("x", "y", "image", "size", "rotation"),
    "triangle": ("x", "y", "x2", "y2", "x3", "y3"),
}

ARG_ALIASES = {
    "x1": "x",
    "y1": "y",
}

def convert_rgb(kwargs: dict) -> dict:
    if "rgb" not in kwargs.keys():
        return kwargs
    color_rgb = int(kwargs["rgb"], 16)
    kwargs["r"] = str((color_rgb >> 16) & 0xFF)
    kwargs["g"] = str((color_rgb >> 8) & 0xFF)
    kwargs["b"] = str((color_rgb >> 0) & 0xFF)
    return kwargs

def draw(src_line: str) -> list:
    """Enhanced draw instruction, with Python-style kwargs.
    Examples:
    # Clear display, using material gray color #373737
    xdraw clear r=55 g=55 b=55
    xdraw clear rgb=0x373737
    # Set stroke width
    xdraw stroke width=1
    # Set color to #FF9100
    xdraw color rgb=0xFF9100
    # Draw a line
    xdraw line x=3 y=1 x2=3 y2=80
    xdraw line x1=3 y1=1 x2=3 y2=80
    # Draw a rectangle
    xdraw rect x1=5 y1=5 height=5 width=10
    # Draw a line rectangle
    xdraw lineRect x1=15 y1=5 height=5 width=10
    # Draw a pentagon
    xdraw poly x=20 y=40 sides=5 radius=10 rotation=0
    # Draw a triangle
    xdraw triangle x1=30 y1=30 x2=20 y2=30 x3=20 y3=20
    # Draw a cyclone
    xdraw color rgb=FFFFFF
    xdraw image x=60 y=60 image=@cyclone size=40 rotation=0
    # Flush
    drawflush display1
    """
    verdicts = src_line.split()
    action = verdicts[1]
    kwargs = parse_kwargs(verdicts[2:])
    kwargs = convert_rgb(kwargs)
    result = []

    try:
        if action in ACTION_ALIASES.keys():
            action = ACTION_ALIASES[action]
        if action not in ACTION_ARGS.keys():
            message = F"error: unsupported xdraw action '{action}'"
            raise CompilationError(message)
        kwargs = apply_aliases(kwargs, ARG_ALIASES)

        output_verdicts = []
        output_verdicts.append("draw")
        output_verdicts.append(action)
        for key in ACTION_ARGS[action]:
            output_verdicts.append(kwargs[key])
        output_verdicts = pad_args(output_verdicts, TOTAL_VERDICTS)
        result.append(" ".join(output_verdicts))
    except KeyError as exception:
        name = ""
        if len(exception.args) >= 1:
            name = exception.args[0]
        message = F"error: argument '{name}' is required, but not found"
        raise CompilationError(message) from exception

    return result



from .jump_if import jump_if
from .extended_let import extended_let
from .unit_control import unit_control
from .radars import general_radar
from .unit_locate import unit_locate
from .control import control
from .draw import draw
from .unsafe_call_return import unsafe_call, unsafe_return

def comment(src_line: str) -> list:
    """Returns an empty list."""
    return []


INSTRUCTIONS = {
    # Yes, comments are instructions
    "#": comment,
    "jump-if": jump_if,
    "xlet": extended_let,
    "unit-control": unit_control,
    "unit-radar": general_radar,
    "unit-locate": unit_locate,
    "xradar": general_radar,
    "xcontrol": control,
    "xdraw": draw,
    "__unsafe_call": unsafe_call,
    "__unsafe_return": unsafe_return,
}



from ..compilation_error import CompilationError
from .kwargs_parser import parse_kwargs, apply_aliases

ARGS = ("filter1", "filter2", "filter3", "sort", "from", "order", "output")
FILTER_NAMES = ("filter1", "filter2", "filter3")
ARG_ALIASES = {
    "target": "filter1",
    "orderBy": "sort",
    "asc": "order",
    "result": "output",
}
SOURCE_IS_UNIT = "__SOURCE_IS_UNIT"

def general_radar(src_line: str) -> list:
    """Enhanced unit radar instruction, with Python-style kwargs.
    """
    verdicts = src_line.split()
    kwargs = parse_kwargs(verdicts[1:])
    kwargs = apply_aliases(kwargs, ARG_ALIASES)

    for filter_id in FILTER_NAMES:
        if filter_id not in kwargs.keys():
            kwargs[filter_id] = "any"

    output_verdicts = []
    if "unit" in src_line:
        output_verdicts.append("uradar")
        kwargs["from"] = "0"
    else:
        output_verdicts.append("radar")

    try:
        for arg in ARGS:
            output_verdicts.append(kwargs[arg])
    except KeyError as exception:
        name = exception.args[0]
        message = F"error: argument '{name}' is required, but not found"
        raise CompilationError(message) from exception
    return [" ".join(output_verdicts), ]



"""Python-style kwargs parser module. Used by ucontrol, uradar, ulocate, etc."""

def pad_args(src: list, total_verdicts: int) -> list:
    """Pad extra args with 0."""
    padding_count = max(total_verdicts - len(src), 0)
    result = []
    result.extend(src)
    result.extend(["0",]*padding_count)
    return result

def parse_kwargs(input_verdicts: list) -> dict:
    kwargs = {}
    for kv_string in input_verdicts:
        key, value = kv_string.split("=", 1)
        kwargs[key] = value
    return kwargs

def apply_aliases(kwdict: dict, aliases: dict) -> dict:
    for alias, real in aliases.items():
        if alias in kwdict.keys():
            kwdict[real] = kwdict[alias]
    return kwdict



"""Provide xlet instruction, a replacement of `op`, `set`, `sensor` and `getlink`."""
from .. import CompilationError

# From Mindustry 126.2
BINARY_OPERATORS = {
    "+": "add",
    "-": "sub",
    "*": "mul",
    "/": "div",
    "//": "idiv",
    "%": "mod",
    "**": "pow",
    "==": "equal",
    "!=": "notEqual",
    "&&": "land",
    "<": "lessThan",
    "<=": "lessThanEq",
    ">": "greaterThan",
    ">=": "greaterThanEq",
    "===": "strictEqual",
    "<<": "shl",
    ">>": "shr",
    "|": "or",
    "&": "and",
    "^": "xor"
}

BINARY_ASSIGNERS = {
    "max": "max",
    "min": "min",
    "angle": "angle",
    "len": "len",
    "noise": "noise",
}

UNARY_ASSIGNERS = {
    "not": "not",
    "abs": "abs",
    "log": "log",
    "log10": "log10",
    "sin": "sin",
    "cos": "cos",
    "tan": "tan",
    "floor": "floor",
    "ceil": "ceil",
    "sqrt": "sqrt",
    "rand": "rand",
    # Custom assigners
    "~": "not",
    "ln": "log",
    "lg": "log10"
}

def unary_xlet(verdicts: list) -> list:
    """xlet for unary operators and assignments."""
    line = ""
    if verdicts[2] == "=":
        # variable assignment, 'set' in Mindustry logic
        lvalue, rvalue = verdicts[1], verdicts[3]
        line = F"set {lvalue} {rvalue}"

    verdicts[2] = verdicts[2].strip("=")
    if verdicts[2] in UNARY_ASSIGNERS.keys():
        vanilla_assigner = UNARY_ASSIGNERS[verdicts[2]]
        lvalue, rvalue = verdicts[1], verdicts[3]
        line = F"op {vanilla_assigner} {lvalue} {rvalue} 0"
    elif verdicts[2] in BINARY_OPERATORS.keys():
        vanilla_operator = BINARY_OPERATORS[verdicts[2]]
        lvalue, rvalue = verdicts[1], verdicts[3]
        line = F"op {vanilla_operator} {lvalue} {lvalue} {rvalue}"
    elif verdicts[2] == "getlink":
        lvalue, link_id = verdicts[1], verdicts[3]
        line = F"getlink {lvalue} {link_id}"
    if len(line) > 0:
        return [line, ]
    message = F"error: unsupported operator '{verdicts[2]}'"
    if verdicts[2] in BINARY_ASSIGNERS.keys():
        message = F"error: too few argument for '{verdicts[2]}'"
    raise CompilationError(message)

def extended_let(src_line: str) -> list:
    """
    C-sytle operators for Mindustry logic. Use '**' for pow/exponetia.
    Usage:
    xlet a = b
    xlet a1 = b + c
    xlet a1 += 3
    xlet a1 = 2 ** 8
    xlet a2 min c d
    xlet a2 =min c d
    xlet a2 =~ x
    xlet a2 floor x
    xlet a3 = b / c
    xlet a4 = b // c
    xlet x =sensor @unit @x
    xlet building =getlink 1
    """
    verdicts = src_line.split()
    if len(verdicts) == 2:
        return []
    if len(verdicts) == 4:
        result = unary_xlet(verdicts)
        return result
    if len(verdicts) == 5:
        verdicts[2] = verdicts[2].lstrip("=")
        line = ""
        if verdicts[2] in BINARY_ASSIGNERS.keys():
            vanilla_assigner = BINARY_ASSIGNERS[verdicts[2]]
            lvalue, arg1, arg2 = verdicts[1], verdicts[3], verdicts[4]
            line = F"op {vanilla_assigner} {lvalue} {arg1} {arg2}"
        elif verdicts[2] == "sensor":
            lvalue, target, attribute = verdicts[1], verdicts[3], verdicts[4]
            line = F"sensor {lvalue} {target} {attribute}"
        if len(line) > 0:
            return [line, ]
        message = F"error: unsupported operator '{verdicts[2]}'"
        raise CompilationError(message)

    if len(verdicts) == 6:
        if verdicts[4] in BINARY_OPERATORS.keys():
            vanilla_operator = BINARY_OPERATORS[verdicts[4]]
            lvalue, arg1, arg2 = verdicts[1], verdicts[3], verdicts[5]
            line = F"op {vanilla_operator} {lvalue} {arg1} {arg2}"
            return [line, ]

        message = F"error: unsupported binary operator '{verdicts[4]}'"
        raise CompilationError(message)

    message = F"error: too {'few' if len(verdicts)<6 else 'many'}"
    message += F"arguments ({len(verdicts)}) for xlet"
    raise CompilationError(message)
